
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\ ~
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\ ~
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\ ~
 //                                                       //            _\|/_
//=======================================================//              /|\ ~

/*
 *  dyn.h: base class for nbody systems
 *.............................................................................
 *    version 1:  Dec 1992   Piet Hut, Steve McMillan, Jun Makino
 *    version 2:
 *.............................................................................
 *     This file includes:
 *  1) definition of class dyn
 *.............................................................................
 */

#ifndef  STARLAB_DYN_H
#  define  STARLAB_DYN_H

#include  "util_math.h"
#include  "node.h"
#include  "kepler.h"

#define BIN_INDENT	21	// for print_pert() and print_binary_params()

//#include  "dyn_kepler.h"   NOTE: this file is included at the end instead,
//                           because it needs to know about the dyn class

/*-----------------------------------------------------------------------------
 *  dyn  --  the simplest class of dynamical particles
 *-----------------------------------------------------------------------------
 */
class  dyn : public node
{
    protected:

        // Global variables:

    	static xreal system_time;	// extended-precision time
    	static real real_system_time;	// convenient real copy

    	static bool use_sstar;  	// activate single star evolution

	// Flag to neglect all internal interactions:

	static bool ignore_internal;

	//------------------------------------------------------------

	// External field specification (moved from hdyn to dyn
	// by Steve, 7/01):

	static unsigned int external_field;
					// 0 ==> none
					// 1 ==> tidal (bit 0)
					// 2 ==> Plummer (bit 1)
					// 3 ==> Plummer+tidal (bits 0+1)
					// 4 ==> power-law (bit 2)
					// etc.

	// Tidal field specifications (changed by Steve, 6/99):
	// Note that alpha and omega are in general independent.

	static int  tidal_type;		// 0 ==> no tidal field
					// 1 ==> point-mass field
					// 2 ==> isothermal halo field
					// 3 ==> disk field (Oort constants)

        static real omega;		// system (circular) angular speed
        static real omega_sq;		// omega squared (probably not needed)

	static real alpha1;		// tidal field is conventionally taken
	static real alpha3;		// to be (-alpha1*x, 0, -alpha3*z)

	static vector tidal_center;	// (fixed) point relative to which
					// tidal forces are computed (7/01)

	// Confining Plummer model parameters (Steve, 7/01):

	static real p_mass;		// total mass
	static real p_scale_sq;		// scale radius squared

	static vector p_center;		// (fixed) point relative to which
					// Plummer forces are computed

	// Confining power-law model parameters (Steve, 7/01):
	// *** Note that Plummer is a special case (beta = 0). ***

	static real pl_coeff;		// overall coefficient
	static real pl_scale_sq;	// scale radius squared
	static real pl_exponent;	// power-law exponent

	static vector pl_center;	// (fixed) point relative to which
					// power-law forces are computed

	// Extra power-law variables added by Steve (12/01):

	static real pl_cutoff;		// inner cutoff radius
	static real pl_cutoff_sq;	// inner cutoff radius squared
	static real pl_mass;		// central point mass
	static real pl_softening_sq;	// softening for point-mass field

	//------------------------------------------------------------

	vector  pos;            // position (3-D Cartesian vector)
	vector  vel;            // velocity: (d/dt) pos
	vector  acc;            // acceleration: (d/dt) vel
	kepler * kep;		// actually a pointer to a kepler orbit object

    public:

	inline void dyn_init() {
	    pos = vel = acc = 0;
	    kep = NULL;
	}

        dyn(hbpfp the_hbpfp = new_hydrobase, sbpfp the_sbpfp = new_starbase,
	    bool use_stories = true)
	   : node(the_hbpfp, the_sbpfp, use_stories)	{dyn_init();}
 
	inline void rmkepler() {
	    if (kep) {
		dyn* s = get_binary_sister();

		if (s && s->kep == kep)
		    s->kep = NULL;

		if ((unsigned int) kep != 1 && (unsigned int) kep != 2)
		    delete kep;		// in case kep is used as a flag

		kep = NULL;
	    }
	}

	virtual ~dyn() {

	    // Note added by Steve 7/7/98:
	    //
	    // Some care is required when deleting the kepler structure,
	    // as binary components in kira share a kepler.  In that case,
	    // when the kepler of the first component is deleted, the
	    // kepler pointer of the other component must also be set
	    // NULL.  Otherwise, an attempt will be made to delete a
	    // nonexistent structure.  Do this via another function so
	    // that the same procedure can be used from other classes.

	    rmkepler();
	}

	inline xreal get_system_time()		{return system_time;}
	inline real get_real_system_time()	{return real_system_time;}
	void set_system_time(xreal t)		{system_time = t;
					         real_system_time = t;}

	void set_ignore_internal(bool i = true)	{ignore_internal = i;}
	bool get_ignore_internal()		{return ignore_internal;}

	//-----------------------------------------------------------------
	// External field:
  
	inline unsigned int get_external_field()       {return external_field;}
	inline void set_external_field(unsigned int e) {external_field = e;}

	// Tidal field (external field #1, bit 0):

	inline void set_tidal_field(int t)	{tidal_type = t;
					         if (t > 0)
						     SETBIT(external_field, 0);
					         else
					     	     CLRBIT(external_field, 0);}
	inline int get_tidal_field()		{return tidal_type;}
	inline void unset_tidal_field(int t)	{set_tidal_field(0);}

        void set_omega(real o)			{omega = o;
						 omega_sq = o*o;}
        inline real get_omega()			{return omega;}
        inline real get_omega_sq()		{return omega_sq;}

	void set_alpha(real a1, real a3)	{alpha1 = a1;
						 alpha3 = a3;}
	inline real get_alpha1()		{return alpha1;}
	inline real get_alpha3()		{return alpha3;}
	inline void set_tidal_center(vector c)	{tidal_center = c;}
	inline vector get_tidal_center()	{return tidal_center;}

	void set_tidal_field(real alpha1, real alpha3,
			     int type = 0, real omega = 0, vector c = 0) {
	    if (type > 0) set_tidal_field(type);
	    set_alpha(alpha1, alpha3);
	    set_tidal_center(c);
	}

	// Plummer field (external field #2, bit 1):

	inline void set_plummer()		{SETBIT(external_field, 1);}
	inline bool get_plummer()		{return (GETBIT(external_field,
								1) > 0);}
	inline void unset_plummer()		{CLRBIT(external_field, 1);}

	inline real get_p_mass()		{return p_mass;}
	inline void set_p_mass(real m)		{p_mass = m;}
	inline real get_p_scale_sq()		{return p_scale_sq;}
	inline void set_p_scale_sq(real r2)	{p_scale_sq = r2;}
	inline void set_p_center(vector c) 	{p_center = c;}
	inline vector get_p_center()		{return p_center;}

	inline void set_plummer(real m, real r2, vector c = 0) {  // all in one
	    set_plummer();
	    p_mass = m;
	    p_scale_sq = r2;
	    p_center = c;
	}

	// Power-law field (external field #3, bit 2):

	inline void set_pl()			{SETBIT(external_field, 2);}
	inline bool get_pl()			{return (GETBIT(external_field,
								2) > 0);}
	inline void unset_pl()			{CLRBIT(external_field, 2);}

	inline real get_pl_coeff()		{return pl_coeff;}
	inline void set_pl_coeff(real c)	{pl_coeff = c;}
	inline real get_pl_scale_sq()		{return pl_scale_sq;}
	inline void set_pl_scale_sq(real r2)	{pl_scale_sq = r2;}
	inline real get_pl_exponent()		{return pl_exponent;}
	inline void set_pl_exponent(real e)	{pl_exponent = e;}
	inline void set_pl_center(vector c) 	{pl_center = c;}
	inline vector get_pl_center()		{return pl_center;}

	inline real get_pl_cutoff()		{return pl_cutoff;}
	inline real get_pl_cutoff_sq()		{return pl_cutoff_sq;}
	inline void set_pl_cutoff(real c)	{pl_cutoff = c;
					         pl_cutoff_sq = c*c;}
	inline real get_pl_mass()		{return pl_mass;}
	inline void set_pl_mass(real m)		{pl_mass = m;}
	inline real get_pl_softening_sq()	{return pl_softening_sq;}
	inline void set_pl_softening_sq(real s)	{pl_softening_sq = s;}

	inline void set_pl(real A, real r2,
			   real e, vector c = 0,
			   real cutoff = 0,
			   real mass = 0,
			   real soft = 0) {	  // all in one

	    set_pl();

	    pl_coeff = A;
	    pl_scale_sq = r2;
	    pl_exponent = e;

	    pl_center = c;

	    set_pl_cutoff(cutoff);
	    pl_mass = mass;
	    pl_softening_sq = soft*soft;
	}

	// Generic accessors for (single) external fields (see dyn_external.C):

	real get_external_scale_sq();
	vector get_external_center();

	//-----------------------------------------------------------------

        void  set_pos(const vector& new_pos)      {pos = new_pos;}
        void  set_vel(const vector& new_vel)      {vel = new_vel;}
        void  set_acc(const vector& new_acc)      {acc = new_acc;}

	void  clear_pos()                         {pos = 0.0;}
	void  clear_vel()                         {vel = 0.0;}
	void  clear_acc()                         {acc = 0.0;}

	inline void  inc_pos(const vector& d_pos) {pos += d_pos;}
	inline void  inc_vel(const vector& d_vel) {vel += d_vel;}
	inline void  inc_acc(const vector& d_acc) {acc += d_acc;}

	inline void  scale_pos(const real scale_factor)  {pos *= scale_factor;}
	inline void  scale_vel(const real scale_factor)  {vel *= scale_factor;}
	inline void  scale_acc(const real scale_factor)  {acc *= scale_factor;}

	inline vector  get_pos()                  {return pos;}
	inline vector  get_vel()                  {return vel;}
	inline vector  get_acc()                  {return acc;}

	inline dyn * get_parent()
	    {return (dyn*) node::get_parent();}
	inline dyn * get_oldest_daughter()
	    {return (dyn*)node::get_oldest_daughter();}
	inline dyn * get_younger_sister()
	    {return (dyn*) node::get_younger_sister();}
	inline dyn * get_elder_sister()
	    {return (dyn*) node::get_elder_sister();}

        inline dyn * get_root()
            {return (dyn*) node::get_root();}
        inline dyn * get_top_level_node()
            {return (dyn*) node::get_top_level_node();}
        inline dyn * get_binary_sister()
            {return (dyn*) node::get_binary_sister();}

	void  calculate_acceleration(dyn *, real);

	inline kepler * get_kepler()		    {return kep;}
	void  set_kepler(kepler * new_kep)	    {kep = new_kep;}

	virtual void null_pointers();
	virtual void print_static(ostream &s = cerr);

        virtual istream& scan_dyn_story(istream&);
	virtual ostream& print_dyn_story(ostream &s,
					 bool print_xreal = true,
					 int short_output = 0);

        void  to_com();      // transformation to center-of-mass coordinates
	void set_com(vector r = 0, vector v = 0);
        void flatten_node();

	// Declaration of member function defined in dyn_tt.C

	virtual real get_radius();
	virtual void set_radius(real) {};

        bool get_use_sstar()			{return use_sstar;}
	void set_use_sstar(bool u)		{use_sstar = u;}

	// Placeholders (~null in dyn, realized in hdyn)

	virtual bool nn_stats(real energy_cutoff, real kT,
			      vector center, bool verbose,
			      bool long_binary_output = true,
			      int which = 0);
	virtual real print_pert(bool long_binary_output = true,
				int indent = BIN_INDENT);
};

typedef dyn * dynptr;  // to enable dynamic array declarations such as
                       //    dyn** dyn_ptr_array = new dynptr[n];
                       // (note that the following expression is illegal:
                       //    dyn** dyn_ptr_array = new (dyn *)[n];)

typedef vector (dyn::*dyn_VMF_ptr)(void);     // vector member function pointer
typedef void (dyn::*dyn_MF_ptr)(const vector &);     // member function pointer

inline  node * new_dyn(hbpfp the_hbpfp, sbpfp the_sbpfp,
		       bool use_stories = true)
    {return (node *) new dyn(the_hbpfp, the_sbpfp, use_stories);}

inline  dyn * mkdyn(int n, hbpfp the_hbpfp = new_hydrobase,
	                   sbpfp the_sbpfp = new_starbase)
    {return  (dyn *) mk_flat_tree(n, new_dyn, the_hbpfp, the_sbpfp);}

inline  dyn * get_dyn(istream & s, hbpfp the_hbpfp = new_hydrobase,
		                   sbpfp the_sbpfp = new_starbase,
		      		   bool use_stories = true)
    {return  (dyn *) get_node(s, new_dyn, the_hbpfp, the_sbpfp, use_stories);}
    
inline dyn * common_ancestor(dyn * bi, dyn * bj)
    {return (dyn*) common_ancestor((node*)bi, (node*)bj);}

void dbg_message(char*, dyn*);
void dbg_message(char*, dyn*, dyn*);

vector something_relative_to_root(dyn*, dyn_VMF_ptr);

// From dyn/init:

void  mksphere(dyn * root, int n, int u_flag = 0);

// From dyn/util:

real pot_on_general_node(dyn * bj, dyn * bi, real eps2, bool cm);
void calculate_energies(dyn * root, real eps2,
			real & epot, real & ekin, real & etot,
			bool cm = false);
void print_recalculated_energies(dyn *, int, real, real e_corr = 0);

void compute_density(dyn* b,
		     int k = 12,
		     dyn** list = NULL,
		     int n_list = 0);

void merge_low_level_nodes(dyn* b, real frac = 1, int option = 1);

// Overloaded functions:

void compute_com(dyn*, vector&, vector&);
void compute_com(dyn*);

void compute_mcom(dyn*, vector&, vector&, real f = 0.9, int n_iter = 2);
void compute_mcom(dyn*, real f = 0.9, int n_iter = 2);

int  get_std_center(dyn*, vector&, vector&, bool verbose = false);
int  get_std_center(dyn*, bool verbose = false);

void compute_max_cod(dyn*, vector&, vector&);
void compute_max_cod(dyn*);

void compute_mean_cod(dyn*, vector&, vector&);
void compute_mean_cod(dyn*);

// From lagrad.C:

void compute_mass_radii(dyn*);
void compute_mass_radii_percentiles(dyn*);
void compute_mass_radii_quartiles(dyn*);

typedef bool boolfn(dyn*);
void compute_general_mass_radii(dyn*, int,
				bool nonlin = false,
				boolfn *bf = NULL);

// From sys_stats.C:

bool parse_sys_stats_main(int argc, char *argv[],
			  int  &which_lagr,
			  bool &binaries,
			  bool &short_output,
			  bool &B_flag,
			  bool &calc_e,
			  bool &n_sq,
			  bool &out,
			  bool &verbose);
void check_addstar(dyn* b);
void sys_stats(dyn* root,
	       real energy_cutoff = 1,
	       bool verbose = true,
	       bool binaries = true,
	       bool long_binary_output = false,
	       int which_lagr = 0,
	       bool print_time = false,
	       bool compute_energy = false,
	       bool allow_n_sq_ops = false,
	       void (*compute_energies)(dyn*, real, real&, real&, real&, bool)
			= calculate_energies,
	       void (*dstar_params)(dyn*) = NULL,
	       bool (*print_dstar_stats)(dyn*, bool, vector, bool) = NULL);

void refine_cluster_mass(dyn *b, int verbose = 0);
void refine_cluster_mass2(dyn *b, int verbose = 0);

// From dyn_stats.C:

real print_binary_params(kepler* k, real m1, real kT,
			 real dist_from_cod,
			 bool verbose = true,
			 bool long_output = true,
			 int init_indent = 0,
			 int indent = BIN_INDENT);

real get_total_energy(dyn* bi, dyn* bj);
real get_period(dyn* bi, dyn* bj);
void get_total_energy_and_period(dyn* bi, dyn* bj, real& E, real& P);

void initialize_kepler_from_dyn_pair(kepler& k, dyn* bi, dyn* bj,
				     bool minimal = false);

void print_binary_from_dyn_pair(dyn* bi, dyn* bj,
				real kT = 0,
				vector center = vector(0,0,0),
				bool verbose = true,
				bool short_output = false);

real print_structure_recursive(dyn* b,
			       void (*dstar_params)(dyn*),
			       int& n_unp, real& e_unp,
			       real kT = 0.0,
			       vector center = vector(0,0,0),
			       bool verbose = true,
			       bool short_output = false,
			       int indent = 0);

real print_structure_recursive(dyn* b,
			       real kT = 0.0,
			       vector center = vector(0,0,0),
			       bool verbose = true,
			       bool short_output = false,
			       int indent = 0);

void compute_core_parameters(dyn*, int, bool, vector&, real&, int&, real&);

// From plot_stars.C:

void plot_stars(dyn * bi,
		int n = 5,
		int k = 3);

// From scale.C:

real get_mass(dyn *b);
void scale_mass(dyn *b, real mscale);
void scale_pos(dyn *b, real rscale, vector com_pos = 0);
void scale_vel(dyn *b, real vscale, vector com_vel = 0);
real get_top_level_kinetic_energy(dyn *b);
real get_kinetic_energy(dyn *b);
void get_top_level_energies(dyn *b, real eps2, real &potential, real &kinetic);
void scale_virial(dyn *b, real q, real potential, real& kinetic,
		  vector com_vel = 0);
real scale_energy(dyn *b, real e, real& energy,
		  vector com_pos = 0,
		  vector com_vel = 0);
bool parse_scale_main(int argc, char *argv[],
		      real& eps, bool& c_flag,
		      bool& e_flag, real& e,
		      bool& m_flag, real& m,
		      bool& q_flag, real& q,
		      bool& r_flag, real& r,
		      bool& debug);
void scale(dyn *b, real eps,
	   bool c_flag,
	   bool e_flag, real e,
	   bool m_flag, real m,
	   bool q_flag, real q,
	   bool r_flag, real r,
	   bool debug = false,
	   void (*top_level_energies)(dyn*, real, real&, real&)
			= get_top_level_energies);

// From dyn_external.C:

void get_external_acc(dyn * b,
		      vector pos,
		      vector vel,
		      real& pot,
		      vector& acc,
		      vector& jerk,
		      bool pot_only = false);
real get_external_pot(dyn * b,
		      void (*pot_func)(dyn *, real) = NULL);

real vcirc(dyn *b, vector r);

real get_tidal_pot(dyn *b);
real get_plummer_pot(dyn *b);
real get_power_law_pot(dyn *b);
real get_external_virial(dyn * b);
void print_external(unsigned int ext, bool shortbits = false);

void set_friction_beta(real b);
void set_friction_mass(real m);
void set_friction_vel(vector v);
void set_friction_acc(dyn *b, real r);

// From add_plummer.C and add_power_law.C:

bool get_physical_scales(dyn *b, real& mass, real& length, real& time);

void add_plummer(dyn *b,
		 real coeff, real scale,
		 vector center = 0,
		 bool n_flag = false,
		 bool verbose = false);

void add_power_law(dyn *b,
		   real coeff, real exponent, real scale,
		   vector center = 0,
		   bool n_flag = false,
		   bool verbose = false,
		   bool G_flag = false);

// From dyn_story.C:

real get_initial_mass(dyn* b,
		      bool verbose = false,
		      bool mass_set = false,
		      real input_mass = 0);
real get_initial_virial_radius(dyn* b,
			       bool verbose = false,
			       bool r_virial_set = false,
			       real input_r_virial = 0);
real get_initial_jacobi_radius(dyn* b,
			       real r_virial,
			       bool verbose = false,
			       bool r_jacobi_set = false,
			       real input_r_jacobi = 0);
void set_tidal_params(dyn* b,
		      bool verbose,
		      real initial_r_jacobi,
		      real initial_mass,
		      int& tidal_field_type);
void test_tidal_params(dyn* b,
		       bool verbose,
		       real initial_r_jacobi,
		       real initial_r_virial,
		       real initial_mass);
int check_set_tidal(dyn *b, bool verbose = false);
void check_set_plummer(dyn *b, bool verbose = false);
void check_set_power_law(dyn *b, bool verbose = false);
void check_set_external(dyn *b, bool verbose = false);
void check_set_ignore_internal(dyn *b, bool verbose = false);

#define  put_dyn  put_node

#include  "dyn_kepler.h"

#endif

//=======================================================================//
//  +---------------+        _\|/_        +------------------------------\\ ~
//  |  the end of:  |         /|\         |  inc/dyn.h
//  +---------------+                     +------------------------------//
//========================= STARLAB =====================================\\ ~
